import java.util.ArrayList;
public class Board
{
  public Board()
  {
    board = new Dice[BOARDSIZE][BOARDSIZE];
  }
  private Board(Board other)
  {
    playerCount = other.playerCount;
    aiCount = other.aiCount;
    board = new Dice[BOARDSIZE][BOARDSIZE];
    for (int i = 0; i < BOARDSIZE; i++)
    {
      for (int j = 0; j < BOARDSIZE; j++)
      {
        board[i][j] = other.board[i][j];
      }
    }
  }
  private int playerCount;
  public int getPlayerCount()
  {
    return playerCount;
  }
  private int aiCount;
  public int getAiCount()
  {
    return aiCount;
  }
  private Dice[][] board;
  // The number of squares per side.
  public static final int BOARDSIZE = 5;
  // The index of the left/bottommost square.
  public static final int LEFT_SIDE = 0;
  // The index of the right/topmost square.
  public static final int RIGHT_SIDE = BOARDSIZE-1;
  // The maximum value of a dice.
  public static final int DICE_MAX = 6;
  // Returns a copy of the current game state.
  public Board copy()
  {
    return new Board(this);
  }
  // Returns whether the game is finished, which happens when the board is full.
  public boolean isFinished()
  {
    return playerCount + aiCount == BOARDSIZE*BOARDSIZE;
  }
  // Returns the dice at position (x, y).
  public Dice getDice(int x, int y)
  {
    return board[x][y];
  }
  // Returns the dice at position (x, y).
  public void setDice(Dice d, int x, int y)
  {
    board[x][y] = d;
  }
  // Returns whether 'move' is valid for 'side'.
  public boolean isValid(Move move, Side side)
  {
    if (move == null)
    {
      return false;
    }
    if (move.getDice().getSide() != side)
    {
      return false;
    }
    int x = move.getX();
    int y = move.getY();
    if (board[x][y] != null)
    {
      return false;
    }
    boolean moveIsACapture = move.getCapture() != null;
    // If we can capture, we must capture (and obviously if the move is a capture, capturing must be allowed).
    // The '^' is a XOR, which means 'either must be true, but they should not both be true'.
    if (moveIsACapture ^ canCapture(x, y))
    {
      return false;
    }
    if (moveIsACapture)
    {
      if (!isValidCapture(move))
      {
        return false;
      }
    }
    else
    {
      if (move.getDice().getValue() != 1)
      {
        return false;
      }
    }
    return true;
  }
  // Returns whether a capture is possible at position (x, y)
  public boolean canCapture(int x, int y)
  {
    // Java is so annoying sometimes...
    for (int captureLeft = 0; captureLeft <= 1; captureLeft++)
    {
      if (captureLeft == 0 || (x > LEFT_SIDE && board[x-1][y] != null))
      {
        for (int captureDown = 0; captureDown <= 1; captureDown++)
        {
          if (captureDown == 0 || (y > LEFT_SIDE && board[x][y-1] != null))
          {
            for (int captureRight = 0; captureRight <= 1; captureRight++)
            {
              if (captureRight == 0 || (x < RIGHT_SIDE && board[x+1][y] != null))
              {
                for (int captureTop = 0; captureTop <= 1; captureTop++)
                {
                  if (captureTop == 0 || (y < RIGHT_SIDE && board[x][y+1] != null))
                  {
                    int sum = captureLeft + captureDown + captureRight + captureTop;
                    if (sum >= 2)
                    {
                      int captureSum = 0;
                      if (captureLeft > 0)
                      {
                        captureSum += board[x-1][y].getValue();
                      }
                      if (captureDown > 0)
                      {
                        captureSum += board[x][y-1].getValue();
                      }
                      if (captureRight > 0)
                      {
                        captureSum += board[x+1][y].getValue();
                      }
                      if (captureTop > 0)
                      {
                        captureSum += board[x][y+1].getValue();
                      }
                      if (captureSum <= 6)
                      {
                        return true;
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
    return false;
  }
  // Returns whether the (capture-) move 'move' is allowed.
  public boolean isValidCapture(Move move)
  {
    int x = move.getX();
    int y = move.getY();
    int captureSum = 0;
    Capture capture = move.getCapture();
    if (capture.getCaptureLeft())
    {
      if (board[x-1][y] == null)
      {
        return false;
      }
      captureSum += board[x-1][y].getValue();
    }
    if (capture.getCaptureDown())
    {
      if (board[x][y-1] == null)
      {
        return false;
      }
      captureSum += board[x][y-1].getValue();
    }
    if (capture.getCaptureRight())
    {
      if (board[x+1][y] == null)
      {
        return false;
      }
      captureSum += board[x+1][y].getValue();
    }
    if (capture.getCaptureUp())
    {
      if (board[x][y+1] == null)
      {
        return false;
      }
      captureSum += board[x][y+1].getValue();
    }
    return move.getDice().getValue() == captureSum;
  }
  // Updates the game state using the valid move 'move'.
  public void makeMove(Move move)
  {
    int x = move.getX();
    int y = move.getY();
    Dice dice = move.getDice();
    addDice(x, y, dice);
    Capture capture = move.getCapture();
    if (capture != null)
    {
      if (capture.getCaptureLeft())
      {
        removeDice(x-1, y);
      }
      if (capture.getCaptureDown())
      {
        removeDice(x, y-1);
      }
      if (capture.getCaptureRight())
      {
        removeDice(x+1, y);
      }
      if (capture.getCaptureUp())
      {
        removeDice(x, y+1);
      }
    }
  }
  private void addDice(int x, int y, Dice dice)
  {
    switch (dice.getSide())
    {
      case AI:
        aiCount++;
        break;
      case Player:
        playerCount++;
        break;
    }
    board[x][y] = dice;
  }
  private void removeDice(int x, int y)
  {
    switch (board[x][y].getSide())
    {
      case AI:
        aiCount--;
        break;
      case Player:
        playerCount--;
        break;
    }
    board[x][y] = null;
  }
  public void undoMove(Move move, Dice leftDice, Dice downDice, Dice rightDice, Dice upDice)
  {
    int x = move.getX();
    int y = move.getY();
    removeDice(x, y);
    Capture capture = move.getCapture();
    if (capture != null)
    {
      if (capture.getCaptureLeft())
      {
        addDice(x-1, y, leftDice);
      }
      if (capture.getCaptureDown())
      {
        addDice(x, y-1, downDice);
      }
      if (capture.getCaptureRight())
      {
        addDice(x+1, y, rightDice);
      }
      if (capture.getCaptureUp())
      {
        addDice(x, y+1, upDice);
      }
    }
  }
  // Returns whether the Player side has won the game.
  public boolean playerHasWon()
  {
    return playerCount > aiCount;
  }
  // Returns all moves that are allowed for 'side'.
  public ArrayList<Move> getAllAllowedMoves(Side side)
  {
    ArrayList<Move> result = new ArrayList<Move>();
    
    // Code duplication for speed.
    for (int x = LEFT_SIDE; x <= RIGHT_SIDE; x++)
    {
      for (int y = LEFT_SIDE; y <= RIGHT_SIDE; y++)
      {
        if (board[x][y] == null)
        {
          boolean foundCapture = false;
          for (int captureLeft = 0; captureLeft <= 1; captureLeft++)
          {
            if (captureLeft == 0 || (x > LEFT_SIDE && board[x-1][y] != null))
            {
              for (int captureDown = 0; captureDown <= 1; captureDown++)
              {
                if (captureDown == 0 || (y > LEFT_SIDE && board[x][y-1] != null))
                {
                  for (int captureRight = 0; captureRight <= 1; captureRight++)
                  {
                    if (captureRight == 0 || (x < RIGHT_SIDE && board[x+1][y] != null))
                    {
                      for (int captureTop = 0; captureTop <= 1; captureTop++)
                      {
                        if (captureTop == 0 || (y < RIGHT_SIDE && board[x][y+1] != null))
                        {
                          int sum = captureLeft + captureDown + captureRight + captureTop;
                          if (sum >= 2)
                          {
                            int captureSum = 0;
                            if (captureLeft > 0)
                            {
                              captureSum += board[x-1][y].getValue();
                            }
                            if (captureDown > 0)
                            {
                              captureSum += board[x][y-1].getValue();
                            }
                            if (captureRight > 0)
                            {
                              captureSum += board[x+1][y].getValue();
                            }
                            if (captureTop > 0)
                            {
                              captureSum += board[x][y+1].getValue();
                            }
                            if (captureSum <= 6)
                            {
                              Capture c = new Capture(captureLeft > 0, captureDown > 0, captureRight > 0, captureTop > 0);
                              Dice d = new Dice(captureSum, side);
                              Move m = new Move(d, x, y, c);
                              result.add(m);
                              foundCapture = true;
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
          if (!foundCapture)
          {
            Dice d = new Dice(1, side);
            Move m = new Move(d, x, y, null);
            result.add(m);
          }
        }
      }
    }
    
    return result;
  }
  // Returns all capture moves that are allowed for 'side'.
  public ArrayList<Move> getAllAllowedCaptureMoves(Side side)
  {
    ArrayList<Move> result = new ArrayList<Move>();
    
    // Code duplication for speed.
    for (int x = LEFT_SIDE; x <= RIGHT_SIDE; x++)
    {
      for (int y = LEFT_SIDE; y <= RIGHT_SIDE; y++)
      {
        if (board[x][y] == null)
        {
          for (int captureLeft = 0; captureLeft <= 1; captureLeft++)
          {
            if (captureLeft == 0 || (x > LEFT_SIDE && board[x-1][y] != null))
            {
              for (int captureDown = 0; captureDown <= 1; captureDown++)
              {
                if (captureDown == 0 || (y > LEFT_SIDE && board[x][y-1] != null))
                {
                  for (int captureRight = 0; captureRight <= 1; captureRight++)
                  {
                    if (captureRight == 0 || (x < RIGHT_SIDE && board[x+1][y] != null))
                    {
                      for (int captureTop = 0; captureTop <= 1; captureTop++)
                      {
                        if (captureTop == 0 || (y < RIGHT_SIDE && board[x][y+1] != null))
                        {
                          int sum = captureLeft + captureDown + captureRight + captureTop;
                          if (sum >= 2)
                          {
                            int captureSum = 0;
                            if (captureLeft > 0)
                            {
                              captureSum += board[x-1][y].getValue();
                            }
                            if (captureDown > 0)
                            {
                              captureSum += board[x][y-1].getValue();
                            }
                            if (captureRight > 0)
                            {
                              captureSum += board[x+1][y].getValue();
                            }
                            if (captureTop > 0)
                            {
                              captureSum += board[x][y+1].getValue();
                            }
                            if (captureSum <= 6)
                            {
                              Capture c = new Capture(captureLeft > 0, captureDown > 0, captureRight > 0, captureTop > 0);
                              Dice d = new Dice(captureSum, side);
                              Move m = new Move(d, x, y, c);
                              result.add(m);
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
    
    return result;
  }
  // Returns all non-capture moves that are allowed for 'side'.
  public ArrayList<Move> getAllAllowedNonCaptureMoves(Side side)
  {
    ArrayList<Move> result = new ArrayList<Move>();
    
    // Code duplication for speed.
    for (int x = LEFT_SIDE; x <= RIGHT_SIDE; x++)
    {
      for (int y = LEFT_SIDE; y <= RIGHT_SIDE; y++)
      {
        if (board[x][y] == null)
        {
          if (!canCapture(x, y))
          {
            Dice d = new Dice(1, side);
            Move m = new Move(d, x, y, null);
            result.add(m);
          }
        }
      }
    }
    
    return result;
  }
}